# frozen_string_literal: true
require 'rbconfig'
require 'socket'

module Atatus
  class Metadata
    class SystemInfo
      # @api private
      class HWInfo
        attr_accessor :cpuinfo_cores, :cpuinfo_model, :cpuinfo_mhz,
        :meminfo_total, :host_bootid

        def initialize
          @os = RbConfig::CONFIG['target_os']
        end

        def read!
          if @os =~ /(linux)/i
            read_from_cpuinfo!
            read_from_meminfo!
            read_from_bootid!
          elsif @os =~ /(darwin)/i
            read_cpuinfo_from_sysctl!
            read_meminfo_from_sysctl!
          end
          self
        end

        def self.read!
          new.read!
        end

        # rubocop:disable Metrics/MethodLength
        def cpuinfo
          @cpuinfo ||=
            begin
              cpuinfo = {
                cores: @cpuinfo_cores,
                model: @cpuinfo_model
              }
              cpuinfo[:mhz] = @cpuinfo_mhz unless @cpuinfo_mhz.nil?

              cpuinfo
            end
        end
        # rubocop:enable Metrics/MethodLength

        def meminfo
          @meminfo ||= @meminfo_total
        end

        # rubocop:disable Metrics/MethodLength
        def hostid
          @hostid ||=
            begin
              return Socket.gethostname if @host_bootid.nil?
              @host_bootid
            end
        end
        # rubocop:enable Metrics/MethodLength

        private

        def get_sysctl_value(key)
          `sysctl -n #{key} 2>/dev/null`
        end

        LINUX_CPUINFO_PATH = '/proc/cpuinfo'.freeze
        PROCESSOR_COUNT_REGEX = /processor\s+:\s+([0-9.]+)/.freeze
        MODEL_NAME_REGEX = /model name.+:(.+)/.freeze
        CPU_MHZ_REGEX = /cpu MHz\s+:\s+([0-9.]+)/.freeze

        LINUX_MEMINFO_PATH = '/proc/meminfo'.freeze
        TOTAL_MEMORY_REGEX = /^MemTotal:\s+(\d+)\skB$/.freeze

        LINUX_BOOTID_PATH = '/proc/sys/kernel/random/boot_id'.freeze

        # rubocop:disable Metrics/MethodLength, Metrics/PerceivedComplexity
        # rubocop:disable Metrics/CyclomaticComplexity, Metrics/AbcSize
        def read_from_cpuinfo!
          return unless File.exist?(LINUX_CPUINFO_PATH)
          cpuinfo = File.read(LINUX_CPUINFO_PATH)
          self.cpuinfo_cores = cpuinfo.scan(PROCESSOR_COUNT_REGEX).size
          self.cpuinfo_model = cpuinfo.scan(MODEL_NAME_REGEX).flatten.first.strip
          self.cpuinfo_mhz = cpuinfo.scan(CPU_MHZ_REGEX).flatten.first
        end
        # rubocop:enable Metrics/MethodLength, Metrics/PerceivedComplexity
        # rubocop:enable Metrics/CyclomaticComplexity, Metrics/AbcSize

        # rubocop:disable Metrics/MethodLength, Metrics/PerceivedComplexity
        # rubocop:disable Metrics/CyclomaticComplexity, Metrics/AbcSize
        def read_from_meminfo!
          return unless File.exist?(LINUX_MEMINFO_PATH)
          meminfo = File.read(LINUX_MEMINFO_PATH)
          self.meminfo_total = (meminfo.scan(TOTAL_MEMORY_REGEX).flatten.first.to_i) * 1024 # to bytes
        end
        # rubocop:enable Metrics/MethodLength, Metrics/PerceivedComplexity
        # rubocop:enable Metrics/CyclomaticComplexity, Metrics/AbcSize

        def read_from_bootid!
          return unless File.exist?(LINUX_BOOTID_PATH)
          self.host_bootid = File.read(LINUX_BOOTID_PATH)
          self.host_bootid.strip!
        end

        def read_cpuinfo_from_sysctl!
          self.cpuinfo_cores = get_sysctl_value('hw.logicalcpu_max').to_i
          self.cpuinfo_model = get_sysctl_value('machdep.cpu.brand_string').strip
        end

        def read_meminfo_from_sysctl!
          self.meminfo_total = get_sysctl_value('hw.memsize').to_i
          self.meminfo_total
        end
      end
    end
  end
end
