# Licensed to Elasticsearch B.V. under one or more contributor
# license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright
# ownership. Elasticsearch B.V. licenses this file to you under
# the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# frozen_string_literal: true

module Atatus
  # @api private
  module Spies
    # @api private
    class NetHTTPSpy
      KEY = :__atatus_net_http_disabled
      TYPE = 'ext'
      SUBTYPE = 'net_http'

      class << self
        def disabled=(disabled)
          Thread.current[KEY] = disabled
        end

        def disabled?
          Thread.current[KEY] ||= false
        end

        def disable_in
          self.disabled = true

          begin
            yield
          ensure
            self.disabled = false
          end
        end
      end

      # rubocop:disable Metrics/CyclomaticComplexity
      def install
        if defined?(::Net) && defined?(::Net::HTTP)

          Net::HTTP.class_eval do
            alias request_without_apm request

            def request(req, body = nil, &block)
              unless (transaction = Atatus.current_transaction)
                return request_without_apm(req, body, &block)
              end

              if Atatus::Spies::NetHTTPSpy.disabled?
                return request_without_apm(req, body, &block)
              end

              host = req['host']&.split(':')&.first || address
              method = req.method.to_s.upcase
              path, query = req.path.split('?')

              url = use_ssl? ? +'https://' : +'http://'
              url << host
              url << ":#{port}" if port
              url << path
              url << "?#{query}" if query
              uri = URI(url)

              destination =
                Atatus::Span::Context::Destination.from_uri(uri)

              context =
                Atatus::Span::Context.new(
                  http: { url: uri, method: method },
                  destination: destination
                )

              Atatus.with_span(
                "#{method} #{host}",
                TYPE,
                subtype: SUBTYPE,
                action: method,
                context: context
              ) do |span|
                trace_context = span&.trace_context || transaction.trace_context
                trace_context.apply_headers { |key, value| req[key] = value }

                result = request_without_apm(req, body, &block)

                if (http = span&.context&.http)
                  http.status_code = result.code
                end

                result
              end
            end
          end

        end
      end
      # rubocop:enable Metrics/CyclomaticComplexity
    end

    register 'Net::HTTP', 'net/http', NetHTTPSpy.new
  end
end
